# Copyright 2021 Lawrence Livermore National Security, LLC and other MuyGPyS
# Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""Distance functions

Compute pairwise and crosswise distance tensors for the purposes of kernel 
construction. 

See the following example computing the pairwise and crosswise distances between 
a batch of training data and their nearest neighbors.

Example:
    >>> from MuyGPyS.neighbors import NN_Wrapper
    >>> from MuyGPyS.optimize.batch import sample_batch
    >>> from MuyGPyS.gp.distance import crosswise_distances
    >>> train_features = load_train_features()
    >>> nn_count = 10
    >>> nbrs_lookup = NN_Wrapper(
    ...         train_features, nn_count, nn_method="exact", algorithm="ball_tree"
    ... )
    >>> train_count, _ = train_features.shape
    >>> batch_count = 50
    >>> batch_indices, batch_nn_indices = sample_batch(
    ...         nbrs_lookup, batch_count, train_count
    ... )
    >>> pairwise_dists = pairwise_distances(
    ...         train_features, batch_nn_inidices, metric="l2"
    ... )
    >>> crosswise_dists = crosswise_distances(
    ...         train_features,
    ...         train_features,
    ...         batch_indices,
    ...         batch_nn_indices,
    ...         metric="l2",
    ... )
    )

See also the following example computing the crosswise distances between a
test dataset and their nearest neighors in the training data.

Example:
    >>> from MuyGPyS.neighbors import NN_Wrapper
    >>> from MuyGPyS.gp.distance import crosswise_distances
    >>> train_features = load_train_features()
    >>> test_features = load_test_features()
    >>> nn_count = 10
    >>> nbrs_lookup = NN_Wrapper(
    ...         train_features, nn_count, nn_method="exact", algorithm="ball_tree"
    ... )
    >>> nn_indices, nn_dists = nbrs_lookup.get_nns(test_features)
    >>> test_count, _ = test_features.shape
    >>> indices = np.arange(test_count)
    >>> nn_indices, _ = nbrs_lookup.get_nns(test_features)
    >>> pairwise_dists = pairwise_distances(
    ...         train_features, nn_inidices, metric="l2"
    ... )
    >>> crosswise_dists = crosswise_distances(
    ...         test_features,
    ...         train_features,
    ...         indices,
    ...         nn_indices,
    ...         metric="l2"
    ... )

"""


import numpy as np

from typing import Optional

from sklearn.metrics.pairwise import cosine_similarity


def crosswise_distances(
    data: np.ndarray,
    nn_data: np.ndarray,
    data_indices: np.ndarray,
    nn_indices: np.ndarray,
    metric: Optional[str] = "l2",
) -> np.ndarray:
    """
    Compute a matrix of distances between data and their nearest neighbors.

    Takes full datasets of records of interest `data` and neighbor candidates
    `nn_data` and produces the distances between each element of `data`
    indicated by `data_indices` and each of the nearest neighbors
    in `nn_data` as indicated by the corresponding rows of `nn_indices`. `data`
    and `nn_data` can refer to the same dataset.

    See the following example computing the crosswise distances between a batch
    of training data and their nearest neighbors.

    Args:
        data:
            The data matrix of shape `(data_count, feature_count)` containing
            batch elements.
        nn_data:
            The data matrix of shape `(candidate_count, feature_count)`
            containing the universe of candidate neighbors for the batch
            elements. Might be the same as `data`.
        indices:
            An integral vector of shape `(batch_count,)` containing the indices
            of the batch.
        nn_indices:
            An integral matrix of shape (batch_count, nn_count) listing the
            nearest neighbor indices for the batch of data points.
        metric:
            The name of the metric to use in order to form distances. Supported
            values are `l2`, `F2`, `ip` (inner product, a distance only if
            data is normalized to the unit hypersphere), and `cosine`.

    Returns:
        A matrix of shape `(batch_count, nn_count)` whose rows list the distance
        of the corresponding batch element to each of its nearest neighbors.
    """
    locations = data[data_indices]
    points = nn_data[nn_indices]
    if metric == "l2":
        return _crosswise_l2(locations, points)
    elif metric == "F2":
        return _crosswise_F2(locations, points)
    elif metric == "ip":
        return _crosswise_prods(locations, points)
    elif metric == "cosine":
        return _crosswise_cosine(locations, points)
    else:
        raise ValueError(f"Metric {metric} is not supported!")


def _crosswise_diffs(locations: np.array, points: np.array) -> np.array:
    return np.array(
        [
            [locations[i, :] - points[i, j, :] for j in range(points.shape[1])]
            for i in range(points.shape[0])
        ]
    )


def _crosswise_F2(locations: np.array, points: np.array) -> np.array:
    return np.array(
        [
            [
                _F2(locations[i, :] - points[i, j, :])
                for j in range(points.shape[1])
            ]
            for i in range(points.shape[0])
        ]
    )


def _crosswise_l2(locations: np.array, points: np.array) -> np.array:
    return np.array(
        [
            [
                _l2(locations[i, :] - points[i, j, :])
                for j in range(points.shape[1])
            ]
            for i in range(points.shape[0])
        ]
    )


def _crosswise_prods(locations: np.array, points: np.array) -> np.array:
    return 1 - np.array(
        [
            [locations[i, :] @ points[i, j, :] for j in range(points.shape[1])]
            for i in range(points.shape[0])
        ]
    )


def _crosswise_cosine(locations: np.array, points: np.array) -> np.array:
    return 1 - np.array(
        [
            [
                cosine_similarity(locations[i, :], points[i, j, :])
                for j in range(points.shape[1])
            ]
            for i in range(points.shape[0])
        ]
    )


def pairwise_distances(
    data: np.ndarray,
    nn_indices: np.ndarray,
    metric: Optional[str] = "l2",
) -> np.ndarray:
    """
    Compute a tensor of pairwise distances among sets of nearest neighbors.

    Takes a full dataset of records of interest `data` and produces the
    pairwise distances between the elements indicated by each row of
    `nn_indices`.

    Args:
        data:
            The data matrix of shape `(data_count, feature_count)` containing
            batch elements.
        nn_indices:
            An integral matrix of shape (batch_count, nn_count) listing the
            nearest neighbor indices for the batch of data points.
        metric:
            The name of the metric to use in order to form distances. Supported
            values are `l2`, `F2`, `ip` (inner product, a distance only if
            data is normalized to the unit hypersphere), and `cosine`.

    Returns:
        A tensor of shape `(batch_count, nn_count, nn_count,)` whose latter two
        dimensions contain square matrices containing the pairwise distances
        between the nearest neighbors of the batch elements.
    """
    points = data[nn_indices]
    if metric == "l2":
        diffs = _diffs(points)
        return _l2(diffs)
    elif metric == "F2":
        diffs = _diffs(points)
        return _F2(diffs)
    elif metric == "ip":
        return _prods(points)
    elif metric == "cosine":
        return _cosine(points)
    else:
        raise ValueError(f"Metric {metric} is not supported!")


def _diffs(points: np.array) -> np.array:
    if len(points.shape) == 3:
        return points[:, :, None, :] - points[:, None, :, :]
    elif len(points.shape) == 2:
        return points[:, None, :] - points[None, :, :]
    else:
        raise ValueError(f"points shape {points.shape} is not supported.")


def _F2(diffs: np.array) -> np.array:
    return np.sum(diffs ** 2, axis=-1)


def _l2(diffs: np.array) -> np.array:
    return np.sqrt(_F2(diffs))


def _prods(points: np.array) -> np.array:
    if len(points.shape) == 3:
        return 1 - np.array([mat @ mat.T for mat in points])
    elif len(points.shape) == 2:
        return 1 - points @ points.T
    else:
        raise ValueError(f"points shape {points.shape} is not supported.")


def _cosine(points: np.array) -> np.array:
    if len(points.shape) == 3:
        return 1 - np.array([cosine_similarity(mat) for mat in points])
    elif len(points.shape) == 2:
        return 1 - cosine_similarity(points)
    else:
        raise ValueError(f"points shape {points.shape} is not supported.")



# Copyright 2021 Lawrence Livermore National Security, LLC and other MuyGPyS
# Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

import os
import sys

from setuptools import find_packages, setup
from typing import Dict


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

INSTALL_REQUIRES = [
    "numpy>=1.18.5",
    "scipy==1.4.1",
    "scikit-learn>=0.23.2",
    "absl-py>=0.13.0",
    "pybind11>=2.5.0",
    "hnswlib>=0.5.2",
]

def _get_version() -> str:
    """Returns the package version.

    Adapted from:
    https://github.com/deepmind/dm-haiku/blob/d4807e77b0b03c41467e24a247bed9d1897d336c/setup.py#L22

    Returns:
      Version number.
    """
    path = "MuyGPyS/__init__.py"
    version = "__version__"
    with open(path) as fp:
        for line in fp:
            if line.startswith(version):
                g: Dict = {}
                exec(line, g)  # pylint: disable=exec-used
                return g[version]  # pytype: disable=key-error
        raise ValueError(f"`{version}` not defined in `{path}`.")


setup(
    name="muygps",
    version=_get_version(),
    description="Scalable Approximate Gaussian Process using Local Kriging",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Benjamin W. Priest",
    author_email="priest2@llnl.gov",
    license="MIT",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=INSTALL_REQUIRES,
    url="https://github.com/LLNL/MuyGPyS",
    download_url="https://pypi.org/project/muygpys",
    project_urls={
        "Source Code": "https://github.com/LLNL/MuyGPyS",
        "Documentation": "https://muygpys.readthedocs.io",
        "Bug Tracker": "https://github.com/LLNL/MuyGPyS/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Development Status :: 3 - Alpha",
    ],
    zip_safe=False,
)



# Copyright 2021 Lawrence Livermore National Security, LLC and other MuyGPyS
# Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""KNN lookup management

`MuyGPyS.neighbors.NN_Wrapper` is an api for tasking several KNN libraries with 
the construction of lookup indexes that empower fast training and inference.
The wrapper constructor expects the training features, the number of nearest 
neighbors, and a method string specifying which algorithm to use, as well as any 
additional kwargs used by the methods.
Currently supported implementations include exact KNN using 
`sklearn <https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.NearestNeighbors.html>`_
("exact") and approximate KNN using `hnsw <https://github.com/nmslib/hnswlib>`_
("hnsw").
"""


import hnswlib
import numpy as np

from sklearn.neighbors import NearestNeighbors
from time import perf_counter
from typing import Tuple, Optional

from MuyGPyS.utils import safe_apply


class NN_Wrapper:
    """
    Nearest Neighbors lookup datastructure wrapper.

    Wraps the logic driving nearest neighbor data structure training and
    querying. Currently supports `sklearn.neighbors.NearestNeighbors` for exact
    computation and `hnswlib.Index` for approximate nearest neighbors.

    An example constructing exact and approximate KNN data lookups with k = 10.

    Example:
        >>> from MuyGPyS.neighors import NN_Wrapper
        >>> train_features = load_train_features()
        >>> nn_count = 10
        >>> exact_nbrs_lookup = NN_Wrapper(
        ...         train_features, nn_count, nn_method="exact", algorithm="ball_tree"
        ... )
        >>> approx_nbrs_lookup = NN_Wrapper(
        ...         train_features, nn_count, nn_method="hnsw", space="l2", M=16
        ... )


    Args:
        train:
            The full training data of shape `(train_count, feature_count)` that
            will construct the nearest neighbor query datastructure.
        nn_count:
            The number of nearest neighbors to return in queries.
        nn_method:
            Indicates which nearest neighbor algorithm should be used.
            Currently "exact" indicates `sklearn.neighbors.NearestNeighbors`,
            while "hnsw" indicates `hnswlib.Index`.
        kwargs:
            Additional kwargs used for lookup data structure construction.
            `nn_method="exact"` supports "radius", "algorithm", "leaf_size",
            "metric", "p", "metric_params", and "n_jobs" kwargs.
            `nn_method="hnsw"` supports "space", "ef_construction", "M", and
            "random_seed" kwargs.
    """

    def __init__(
        self,
        train: np.ndarray,
        nn_count: int,
        nn_method: str = "exact",
        **kwargs,
    ):
        """
        NOTE[bwp] Will need to replace `train` with a data stream in the future.
        """
        self.train = train
        self.train_count, self.feature_count = self.train.shape
        self.nn_count = nn_count
        self.nn_method = nn_method.lower()
        if self.nn_method == "exact":
            exact_kwargs = {
                k: kwargs[k]
                for k in kwargs
                if k
                in {
                    "radius",
                    "algorithm",
                    "leaf_size",
                    "metric",
                    "p",
                    "metric_params",
                    "n_jobs",
                }
            }

            exact_kwargs["n_neighbors"] = nn_count + 1
            exact_kwargs["n_jobs"] = exact_kwargs.get("n_jobs", -1)
            self.nbrs = NearestNeighbors(**exact_kwargs).fit(self.train)
        elif self.nn_method == "hnsw":
            self.nbrs = hnswlib.Index(
                space=kwargs.get("space", "l2"), dim=self.feature_count
            )
            hnsw_kwargs = {
                k: kwargs[k]
                for k in kwargs
                if k in {"ef_construction", "M", "random_seed"}
            }
            hnsw_kwargs["max_elements"] = self.train_count
            self.nbrs.init_index(**hnsw_kwargs)
            self.nbrs.add_items(self.train)
        else:
            raise NotImplementedError(
                f"Nearest Neighbor algorithm {self.nn_method} is not implemented."
            )

    def get_nns(
        self,
        test: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get the nearest neighbors for each row of `test` dataset.

        Find the nearest neighbors and associated distances for each element of
        the given test dataset. Here we assume that the test dataset is distinct
        from the train dataset used in the construction of the nearest neighbor
        lookup data structure.

        Example:
            >>> from MuyGPyS.neighbors import NN_Wrapper
            >>> train_features = load_train_features()
            >>> test_features = load_test_features()
            >>> nn_count = 10
            >>> nbrs_lookup = NN_Wrapper(
            ...         train_features, nn_count, nn_method="exact", algorithm="ball_tree"
            ... )
            >>> nn_indices, nn_dists = nbrs_lookup.get_nns(test_features)


        Args:
            test:
                Testing data matrix of shape `(test_count, feature_count)`.

        Returns
        -------
        nn_indices:
            Matrix of nearest neighbor indices of shape
            `(test_count, nn_count)`. Each row lists the nearest neighbor
            indices of the corresponding test element.
        nn_dists:
            Matrix of distances of shape `(test_count, nn_count)`. Each row
            lists the distance to the test element of the corresponding element
            in `nn_indices`.
        """
        return self._get_nns(test, self.nn_count)

    def get_batch_nns(
        self,
        batch_indices: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get the non-self nearest neighbors for indices into the training data.

        Find the nearest neighbors and associated distances for each specified
        index into the training data.

        Example:
            >>> from MuyGPyS.neighbors import NN_Wrapper
            >>> from numpy.random import choice
            >>> train_features = load_train_features()
            >>> nn_count = 10
            >>> nbrs_lookup = NN_Wrapper(
            ...         train_features, nn_count, nn_method="exact", algorithm="ball_tree"
            ... )
            >>> train_count, _ = train_features.shape
            >>> batch_count = 50
            >>> batch_indices = choice(train_count, batch_count, replace=False)
            >>> nn_indices, nn_dists = nbrs_lookup.get_nns(batch_indices)

        Args:
            batch_indices:
                Indices into the training data of shape `(batch_count,)`.

        Returns
        -------
        batch_nn_indices:
            Matrix of nearest neighbor indices of shape
            `(batch_count, nn_count)`. Each row lists the nearest neighbor
            indices (self excluded) of the corresponding batch element.
        batch_nn_dists : numpy.ndarray(int), shape=(batch_count, nn_count)
            Matrix of distances of shape `(batch_count, nn_count)`. Each row
            lists the distance to the batch element of the corresponding element
            in `batch_nn_indices`.
        """
        batch_nn_indices, batch_nn_dists = self._get_nns(
            self.train[batch_indices, :],
            self.nn_count + 1,
        )
        return batch_nn_indices[:, 1:], batch_nn_dists[:, 1:]

    def _get_nns(
        self,
        samples: np.ndarray,
        nn_count: int,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get the nearest neighbors for each row of `samples` dataset.

        Args:
            samples:
                Data matrix of shape `(sample_count, feature_count)` whose rows
                include samples to be queried.
            nn_count:
                The number of nearest neighbors to query.

        Returns
        -------
        nn_indices:
            Matrix of nearest neighbor indices of shape
            `(sample_count, nn_count)`. Each row lists the nearest neighbor
            indices of the corresponding samples element.
        nn_dists:
            Matrix of distances of shape `(sample_count, nn_count)`. Each row
            lists the distance to the sample element of the corresponding
            element in `nn_indices`.
        """
        if self.nn_method == "exact":
            nn_dists, nn_indices = self.nbrs.kneighbors(
                samples,
                n_neighbors=nn_count,
            )
            if self.nbrs.metric == "minkowski" and self.nbrs.p == 2:
                # We do this so that both implementations return the squared l2
                # for downstream consistency. Taking the square root is much
                # more expensive, so this should not produce much overhead.
                nn_dists = nn_dists ** 2
        elif self.nn_method == "hnsw":
            # Although hnsw uses 'l2' as the name of its metric, it returns
            # F2 values as distances in order to avoid the square root
            # computations.
            nn_indices, nn_dists = self.nbrs.knn_query(samples, k=nn_count)
        else:
            raise NotImplementedError(
                f"Nearest Neighbor algorithm {self.nn_method} is not implemented."
            )

        return nn_indices, nn_dists



# Copyright 2021 Lawrence Livermore National Security, LLC and other MuyGPyS
# Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import numpy as np

from typing import Iterable, List, Optional, Tuple, Union

from sklearn.gaussian_process.kernels import Matern, RBF


class BenchmarkGP:
    """
    A basic Gaussian Process.

    Performs GP inference and simulation by way of analytic computations.

    Args:
        kern:
            The kernel to be used. Each kernel supports different
            hyperparameters that can be specified in kwargs.
            NOTE[bwp] Currently supports `matern` and `rbf`.
        **kwargs:
            Kernel parameters. See :ref:`MuyGPyS-gp-kernels`.
    """

    def __init__(
        self,
        kern: str = "matern",
        **kwargs,
    ):
        """
        Initialize.
        """
        self.kern = kern.lower()
        self.set_params(**kwargs)

    def set_params(self, **params) -> List[str]:
        """
        Set the hyperparameters specified by `params`.

        NOTE[bwp] this logic should get moved into kernel functors once
        implemented

        Args:
            eps:
                The homoscedastic noise nugget to be added to the inverted
                covariance matrix.
            sigma_sq:
                Scaling parameter to be applied to posterior variance. One
                element per dimension of the response.
            nu:
                The smoothness parameter. As `nu` -> infty, the matern kernel
                converges pointwise to the RBF kernel. Accept only if this is
                a matern GP.
            length_scale:
                Scale parameter multiplied against distance values. Accept if
                RBF or Matern.

        Returns:
            The set of kernel parameters that have not been fixed by `params`.
        """
        self.params = {
            p: params[p] for p in params if p != "eps" and p != "sigma_sq"
        }
        self.eps = params.get("eps", 0.015)
        self.sigma_sq = params.get("sigma_sq", np.array(1.0))
        if self.kern == "matern":
            self.kernel = Matern(
                length_scale=self.params.get("length_scale", 10.0),
                nu=self.params.get("nu", 0.5),
            )
            unset_params = {"eps", "sigma_sq", "length_scale", "nu"}.difference(
                params.keys()
            )
        elif self.kern == "rbf":
            self.kernel = RBF(length_scale=self.params.get("length_scale", 0.5))
            unset_params = {"eps", "sigma_sq", "length_scale"}.difference(
                params.keys()
            )
        else:
            raise NotImplementedError(f"{self.kern} is not implemented yet!")
        return sorted(list(unset_params))

    def set_param_array(
        self,
        names: List[str],
        values: List[float],
    ) -> None:
        """
        Set the hyperparameters specified by elements of `names` with the
        corresponding elements of `values`.

        Convenience function for use in concert with `scipy.optimize`.

        NOTE[bwp] this logic should get moved into kernel functors once
        implemented

        Args:
            names:
                An alphabetically ordered list of parameter names.
            values:
                A corresponding list of parameter values.
        """
        names = list(names)
        # this is going to break if we add a hyperparameter that occurs earlier
        # in alphabetical order.
        if names[0] == "eps":
            self.eps = values[0]
            names = names[1:]
            values = values[1:]
        for i, name in enumerate(names):
            self.params[name] = values[i]
        if self.kern == "matern":
            self.kernel = Matern(**self.params)
        elif self.kern == "rbf":
            self.kernel = RBF(**self.params)

    def optim_bounds(
        self,
        names: Iterable[str],
        eps: float = 1e-6,
    ) -> List[Tuple[float, float]]:
        """
        Set the bounds (2-tuples) corresponding to each specified
        hyperparameter.

        NOTE[bwp] this logic should get moved into kernel functors once
        implemented

        Args:
            names:
                An iterable over hyperparameter names.
        """
        ret = list()
        if "eps" in names:
            ret.append((eps, 0.2))
        if self.kern == "matern":
            if "length_scale" in names:
                ret.append((eps, 40.0))
            if "nu" in names:
                ret.append((eps, 2.0))
        elif self.kern == "rbf":
            if "length_scale" in names:
                ret.append((eps, 40.0))
        return ret

    def fit(
        self,
        test: np.ndarray,
        train: np.ndarray,
    ) -> None:
        """
        Compute the full kernel and precompute the cholesky decomposition.

        Args:
            test:
                The full testing data matrix of shape
                `(test_count, feature_count)`.
            train:
                The full training data matrix of shape
                `(train_count, feature_count)`.
        """
        self._fit_kernel(np.vstack((test, train)))
        self.test_count = test.shape[0]
        self._cholesky(self.K)

    def fit_train(self, train: np.ndarray) -> None:
        """
        Compute the training kernel and precompute the cholesky decomposition.

        Args:
            train:
                The full training data matrix of shape
                `(train_count, feature_count)`.
        """
        self._fit_kernel(train)
        self.test_count = 0
        self._cholesky(self.K)

    def _fit_kernel(self, x):
        self.K = self.kernel(x) + self.eps * np.eye(x.shape[0])

    def _cholesky(self, K):
        self.cholK = np.linalg.cholesky(K)

    def simulate(self):
        return self.cholK @ np.random.normal(0, 1, size=(self.cholK.shape[0],))

    def get_sigma_sq(self, y):
        assert y.shape[0] == self.K.shape[0]
        return (1 / y.shape[0]) * y @ np.linalg.solve(self.K, y)

    def regress(
        self,
        targets: np.ndarray,
        variance_mode: Optional[str] = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """
        Performs simultaneous regression on a list of observations.

        Args:
            targets:
                A matrix of shape `(train_count, ouput_dim)` whose rows consist
                of vector-valued responses for each training element.
            variance_mode:
                Specifies the type of variance to return. Currently supports
                `diagonal` and None. If None, report no variance term.

        Returns
        -------
        responses:
            A matrix of shape `(batch_count, response_count)` whose rows consist
            of the predicted response for each of the given indices.
        diagonal_variance:
            A vector of shape `(batch_count,)` consisting of the diagonal
            elements of the posterior variance. Only returned where
            `variance_mode == "diagonal"`.
        """
        if self.test_count == 0:
            return np.array([])
        Kcross = self.K[self.test_count :, : self.test_count]
        K = self.K[: self.test_count, : self.test_count]
        responses = Kcross @ np.linalg.solve(K, targets)

        if variance_mode is None:
            return responses
        elif variance_mode == "diagonal":
            Kstar = self.K[self.test_count :, self.test_count :]
            variance = Kstar - Kcross @ np.linalg.solve(K, Kcross.T)
            return responses, np.diagonal(variance)
        elif variance_mode == "full":
            Kstar = self.K[self.test_count :, self.test_count :]
            variance = Kstar - Kcross @ np.linalg.solve(K, Kcross.T)
            return responses, variance
        else:
            raise NotImplementedError(
                f"Variance mode {variance_mode} is not implemented."
            )
